/*
 * DAQmanager.cc
 *
 *  Created on: 29 Jan 2011
 *      Author: lacasta
 */

#include <iostream>
#include <sstream>
#include "utils.h"
#include "DAQmanager.h"
#include <RunManager.h>
#include <DataFile.h>
#include <DataMonitor.h>
#include <AsciiDataFile.h>
#include <HDFDataFile.h>
#include "ConfigFactory.h"
#include "Analysis.h"
#include "DAQConfig.h"
#include "PluginConfig.h"
#include "AlibavaMonitor.h"
#include "utils.h"

DAQmanager *DAQmanager::_mngr=0;


struct DAQmanagerPriv
{
        AlibavaMutex mtx;
};

DAQmanager::DAQmanager(int nn, Alibava *B, RunManager::RunType tp) :
    nevts(nn), received(0), expected(nn), board(B), ofile(0), format(DataFile::HDF5), monitor(0),
    run_type(RunManager::Unknown), rm(0),
    _starting(false), _running(false), _stop_running(false), _is_gui(false), config_file( 0), _priv(0)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    config_file = new KeyFile();
    cfg->set_file(config_file);

    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));

    daq->nevts(nn);
    board->set_data_receiver(this);
    board->delay( daq->delay() );
    set_run_type(tp);

    _priv = new DAQmanagerPriv;
}

DAQmanager::~DAQmanager()
{
    delete _priv;
    if (rm)
        delete rm;

    delete config_file;

}


bool DAQmanager::running() const
{
    bool rc;
    _priv->mtx.lock();
    rc = _running;
    _priv->mtx.unlock();
    return rc;
}
void DAQmanager::running(bool x)
{
    _priv->mtx.lock();
    _running = x;
    _starting = false;
    _priv->mtx.unlock();
}


bool DAQmanager::starting() const
{
    bool rc;
    _priv->mtx.lock();
    rc =  _starting;
    _priv->mtx.unlock();
    return rc;
}

bool DAQmanager::stop() const
{
    bool rc;
    _priv->mtx.lock();
    rc =  _stop_running;
    _priv->mtx.unlock();
    return rc;
}
void DAQmanager::stop(bool x)
{
    _priv->mtx.lock();
    _stop_running = x;
    _priv->mtx.unlock();
}

int DAQmanager::open(const char *fnam)
{
    board->open(fnam);
    if (!(*board))
    {
        err_msg( _("Something failed when opening the device") );
        return 1;
    }

    // reset the card
    if (reset()<0)
        return -1;

    device_opened();
    return 0;
}

int DAQmanager::reset()
{
    int rc;
    write_msg(_("*** Reseting the board\n"));
    // First thing we do is to send a reset
    rc = board->software_reset();
    if (rc)
        return -1;

    return ConfigFactory::instance()->config_board(*board);
}


void DAQmanager::set_run_type( RunManager::RunType tp )
{
    if (tp == RunManager::Unknown || tp == run_type )
        return;

    if (rm)
        delete rm;

    run_type = tp;
    rm = RunManager::factory(tp, this);
    if (rm)
    {
        ConfigFactory *cfg = ConfigFactory::instance();
        PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
        if ( plg->active())
            rm->set_plugin(plg->plugin());
    }
    if (!rm)
    {
        err_msg( _("Could not create an appropriate RunManager\n") );
    }
}

void DAQmanager::set_data_file(const std::string &fnam)
{
    if (!fnam.empty())
    {
        ofile = DataFile::factory(format, fnam);
    }
}

int DAQmanager::prepare_for_run()
{
    ConfigFactory *cfg = ConfigFactory::instance();
    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));
    Alibava *alibava = get_board();

    if (!rm)
        return -1;

    received = 0;
    _starting = true;
    board->stop(true);
    board->delay( daq->delay() );
    ana->set_mask(  alibava->mask() );
    ana->reset(true);
    /*
     * Prepare the output file (if any)
     */
    if (ofile)
    {
        ofile->scan_definition(0, 0, 0, 0, 0);
        ofile->describe_setup(get_run_type(), alibava->nchip(), alibava->mask(), alibava->firmware());
        ofile->open();
        /*
         * Let the plugin write a header to the file
         */
        if (plg->active())
        {
            std::string header;
            std::string data_format;
            plg->plugin()->get_data_format(data_format);
            ofile->set_data_format(data_format.c_str());
            plg->plugin()->new_file(header);
            if (!header.empty())
                ofile->write(DataFile::NewFile, header.size(), header.c_str());
        }
    }


    if ( plg->active())
        rm->set_plugin(plg->plugin());

    rm->start_of_run();

    daq->nevts( nevts );
    if (plg->active())
    {
        int nn = plg->plugin()->start_of_run(run_type, daq->nevts(), daq->sample_size());
        daq->nevts( nn );
    }
    expected = daq->nevts();
    return 0;
}
int DAQmanager::acquire() {
    ConfigFactory *cfg = ConfigFactory::instance();
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));

    std::ostringstream oheader;
    if (ofile)
        ofile->write(run_type,0,0);

    board->stop(false);
    oheader << "V3.0|";

    running(true);
    stop(false);

    int rc = rm->loop(oheader);
    if ( rc >= 0 )
    {
        if ( plg->active() )
        {
            std::string buffer;
            plg->plugin()->end_of_run(buffer);
            if ( ofile )
            {
                if (!buffer.empty())
                    ofile->write(DataFile::EndOfRun, buffer.size(), buffer.c_str());
            }
        }
        write_msg(_("Run finished\n"));

    }
    else
    {
        reset();
    }
    if (ofile)
    {
        std::ostringstream ostr;
        ofile->set_header(oheader.str());
        ofile->write(run_type, ana->pedestals(), ana->noise());
        ofile->close();

        write_msg(_("Data stored in %s\n"), ofile->file_name().c_str());
        delete ofile;
        ofile = 0;
    }
    running(false);
    stop(false);
    return rc;
}

int DAQmanager::stop_run()
{
    stop(true);
    return 0;
}


void DAQmanager::receive(const AlibavaData *b)
{
    received++;
    process_event(b);
}

void DAQmanager::write_event(const AlibavaData *b)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));

    if (plg->active())
    {
        std::string data;
        plg->plugin()->filter_event(b, data);
        if (ofile)
        {
            if (data.empty())
            {
                ofile->write(b);
            }
            else
            {
                ofile->write(DataFile::DataBlock, data.size(), data.c_str());
            }
        }
    }
    else
    {
        // write data
        if (ofile)
            ofile->write(b);
    }
}

void DAQmanager::process_event(const AlibavaData *b)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));
    ana->process_event((unsigned short *)b->chip(0), board->mask());
    write_event(b);
}


void DAQmanager::flush_data_file()
{
    if (ofile)
        ofile->flush();
}

void DAQmanager::load_config(const char *fnam, bool load)
{
    ConfigFactory *cfg = ConfigFactory::instance();

    if (config_file)
        delete config_file;

    config_file = new KeyFile();

    std::ifstream ifile(fnam);
    if (!ifile)
    {
        err_msg( _("could not open config file %s"), fnam);
        return;
    }
    config_file->load(ifile);
    ifile.close();
    cfg->set_file(config_file);

    if (load)
        cfg->config_board( *get_board() );

    config_loaded();
}

void DAQmanager::config_loaded()
{

}

void DAQmanager::save_config(const char *fnam)
{
    ConfigFactory::instance()->save_data();

    std::ofstream ofile(fnam);
    if (!ofile)
    {
        err_msg( _("could not dump config: %s"), fnam );
    }
    else
    {
        config_file->save(ofile);
    }
}



void DAQmanager::set_daq_manager_data_file(const std::string &fnam)
{
    if (_mngr)
    {
        info_msg("Setting data file to %s", fnam.c_str());
        _mngr->set_data_file(fnam);
    }
}

void DAQmanager::daq_manager_stop_run()
{
    if (_mngr )
    {
        info_msg("Stopping run");
        _mngr->stop_run();
    }
}
int DAQmanager::daq_manager_reset_board()
{
    int rc = 0;
    if (_mngr)
        rc = _mngr->reset();
    return rc;
}

gboolean update_monitor(gpointer ptr)
{
    extern AlibavaMonitor __monitor__;

    if (  DAQmanager::get_the_manager()->is_gui() )
        return FALSE;

    __monitor__.computeTimeElapsed();
    __monitor__.computeRate(DAQmanager::get_the_manager()->get_received());

    return TRUE;
}




int DAQmanager::daq_manager_start_run(int run_type, int num_events, int sample_size, DataReceiver *D)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));
    extern AlibavaMonitor __monitor__;
    if ( run_type < 3 )
    {
        daq->nevts(num_events);
        daq->sample_size(sample_size);
    }
    RunManager::RunType the_type;
    int rc;

    switch (run_type)
    {
        case 0:
            the_type = RunManager::Pedestal;
            break;
        case 1:
            the_type = RunManager::RadSource;
            break;
        case 2:
            the_type = RunManager::Laser;
            break;
        case 3:
            the_type = RunManager::LaserSync;
            break;
        case 4:
            the_type = RunManager::Calibration;
            break;
        case 5:
            the_type = RunManager::ChargeScan;
            break;
        default:
            std::cout << "Unknown Run type" << std::endl;
            return -1;
    }
    __monitor__.setRunTypeVal(the_type);
    __monitor__.setNexpected(_mngr->get_expected());
    _mngr->set_run_type(the_type);

    if (D)
        _mngr->get_board()->set_data_receiver(D);

    __monitor__.from_soap(true);
    int tmout = g_timeout_add(500, (GSourceFunc)update_monitor, 0);
    if (_mngr->is_gui())
    {
        rc = _mngr->prepare_for_run();
        if (rc)
        {
            if (D)
                _mngr->get_board()->remove_data_receiver(D);
            return rc;
        }
        __monitor__.isRunning(true);

        _mngr->acquire_thread(false);
    }
    else
    {
        _mngr->prepare_for_run();
        __monitor__.isRunning(true);
        _mngr->acquire();
    }
    g_source_remove(tmout);
    if (D)
        _mngr->get_board()->remove_data_receiver(D);

    _mngr->stop_run();
    __monitor__.isRunning(false);


    return 0;
}


