#include <iostream>
#include <iomanip>
#include <algorithm>
#include <string>
#include <sstream>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <sys/types.h>
#include <unistd.h>
#include "KeyFile.h"
#include <utils.h>
#include <glib.h>


struct KeyFilePriv
{
        GKeyFile *kfile;

};



// check if a current group exists
bool KeyFile::has_group(const std::string&s) const
{
    return g_key_file_has_group(priv->kfile, s.c_str());
}

// check if a given group has a given key
bool KeyFile::has_key(const std::string &g, const std::string &k) const
{
    GError *err=0;
    bool rc = g_key_file_has_key(priv->kfile, g.c_str(), k.c_str(), &err);
    if (err)
    {
        err_msg("Error while querying for existence of [%s].%s\n---> %s", g.c_str(), k.c_str(), err->message);
        g_error_free(err);
    }
    return rc;
}

int KeyFile::add_group(const std::string &g)
{
    return 0;
}

void KeyFile::add_key(const std::string &g, const std::string &k, const std::string &val)
{
    g_key_file_set_string(priv->kfile, g.c_str(), k.c_str(), val.c_str());
}

void KeyFile::del_key(const std::string &g, const std::string &k)
{
    g_key_file_remove_key(priv->kfile, g.c_str(), k.c_str(), NULL);
}

void KeyFile::del_group(const std::string &g)
{
    g_key_file_remove_group(priv->kfile, g.c_str(), NULL);
}


KeyFile::Exception::Exception()
{
}

KeyFile::Exception::Exception(const KeyFile::Exception &e)
{
    _why = e._why;
}

KeyFile::Exception::~Exception() throw()
{
}

static char buffer[2048];

const char* KeyFile::Exception::what() const throw()
{
    std::string rc;
    if (_why.empty() )
        rc = "KeyFile::Exception";
    else
        rc = _why;

    sprintf(buffer,"process (%d) %s", getpid(), rc.c_str());
    return buffer;
}


KeyFile::KeyFile()
{
    priv = new KeyFilePriv;
    priv->kfile = g_key_file_new();
}

KeyFile::~KeyFile()
{
    reset();
    g_key_file_free(priv->kfile);
    delete priv;
}

void KeyFile::reset()
{
    g_key_file_free(priv->kfile);

    priv->kfile = g_key_file_new();
}

std::string KeyFile::value(const std::string &g, const std::string &k) const throw(Exception)
{
    GError *err = 0;
    std::string rc("");
    const char *val = g_key_file_get_value(priv->kfile, g.c_str(), k.c_str(), &err);
    if (err)
    {
        err_msg( _("Error getting the value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        return rc;
    }
    rc = val;
    return rc;
}

int KeyFile::load(std::istream &is)
{
    GError *err = 0;
    std::stringstream buffer;
    buffer << is.rdbuf();
    std::string xx = buffer.str();

    g_key_file_load_from_data(priv->kfile,
                          buffer.str().c_str(),
                          buffer.str().length(),
                          (GKeyFileFlags)(G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS),
                          &err);
    if (err)
    {
        err_msg(_("Error reading config file: %s"), err->message );
        g_error_free(err);
        return -1;
    }

    return 0;
}
int KeyFile::save(std::ostream &os) const
{
    gsize lenth=0;
    char *out = g_key_file_to_data(priv->kfile, &lenth, NULL);

    os.write(out, lenth);
    return 0;
}

std::ostream &operator<<(std::ostream &os, const KeyFile &kf)
{
    kf.save(os);
    return os;
}

std::istream &operator>>(std::istream &is, KeyFile &kf)
{
    kf.load(is);
    return is;
}

bool KeyFile::get_as_bool(const std::string &g, const std::string &k) const throw(KeyFile::Exception)
{
    GError *err = 0;
    gboolean rc = g_key_file_get_boolean(priv->kfile, g.c_str(), k.c_str(), &err);
    if (err)
    {
        std::string errmsg(err->message);
        err_msg(_("Error retrieveing value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        throw KeyFile::Exception(err->message);
    }
    return rc;
}

void KeyFile::get_as_bool_list(const std::string &g, const std::string &k, std::vector<bool> &v) throw(KeyFile::Exception)
{
    GError *err=0;
    gsize length = 0;
    gboolean *buff = g_key_file_get_boolean_list(priv->kfile, g.c_str(), k.c_str(), &length, &err);
    if (err)
    {
        std::string errmsg(err->message);
        err_msg(_("Error retrieveing value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        throw KeyFile::Exception(err->message);
    }
    for (gsize i = 0; i<length; i++)
    {
        v.push_back(buff[i]);
    }
    g_free(buff);
}

void KeyFile::set_as_bool_list(const std::string &g, const std::string &k, const std::vector<bool> &v)
{
    gsize length = v.size();
    gboolean *buf = new gboolean[length];
    for (gsize i=0; i<length; i++)
        buf[i] = v[i];

    g_key_file_set_boolean_list(priv->kfile, g.c_str(), k.c_str(),buf, length);
    delete [] buf;
}

void KeyFile::set_as_bool(const std::string &g, const std::string &k, bool x)
{
    g_key_file_set_boolean(priv->kfile, g.c_str(), k.c_str(), x);
}

double KeyFile::get_as_double(const std::string &g, const std::string &k) const throw(KeyFile::Exception)
{
    GError *err = 0;
    double x = g_key_file_get_double(priv->kfile, g.c_str(), k.c_str(), &err);
    if (err)
    {
        std::string errmsg(err->message);
        err_msg(_("Error retrieveing value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        throw KeyFile::Exception(err->message);
    }
    return x;
}

void KeyFile::get_as_double_list(const std::string &g, const std::string &k, std::vector<double> &v) throw(KeyFile::Exception)
{
    GError *err = 0;
    gsize length = v.size();
    double *rc = g_key_file_get_double_list(priv->kfile, g.c_str(), k.c_str(), &length, &err);
    if (err)
    {
        std::string errmsg(err->message);
        err_msg(_("Error retrieveing value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        throw KeyFile::Exception(err->message);
    }
    for (gsize i=0; i<length; i++)
        v.push_back(rc[i]);

    g_free(rc);
}

void KeyFile::set_as_double(const std::string &g, const std::string &k, double val)
{
    g_key_file_set_double(priv->kfile, g.c_str(), k.c_str(), val);
}
void KeyFile::set_as_double_list(const std::string &g, const std::string &k, const std::vector<double> &v)
{
    if (v.empty())
        return;

    gsize length = v.size();
    double *buf = new double[length];
    for (gsize i=0; i<length; i++)
        buf[i] = v[i];

    g_key_file_set_double_list(priv->kfile, g.c_str(), k.c_str(),buf, length);
    delete [] buf;
}

int KeyFile::get_as_int(const std::string &g, const std::string &k) const throw(KeyFile::Exception)
{
    GError *err = 0;
    gint ival = g_key_file_get_integer(priv->kfile, g.c_str(), k.c_str(), &err);
    if (err)
    {
        std::string errmsg(err->message);
        err_msg(_("Error retrieveing value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        throw KeyFile::Exception(err->message);
    }
    return ival;

}
void KeyFile::get_as_int_list(const std::string &g, const std::string &k, std::vector<int> &v) throw(KeyFile::Exception)
{
    GError *err = 0;
    gsize length = v.size();
    gint *rc = g_key_file_get_integer_list(priv->kfile, g.c_str(), k.c_str(), &length, &err);
    if (err)
    {
        std::string errmsg(err->message);
        err_msg(_("Error retrieveing value of [%s].%s: %s"), g.c_str(), k.c_str(), err->message);
        g_error_free(err);
        throw KeyFile::Exception(err->message);
    }
    for (gsize i=0; i<length; i++)
        v.push_back(rc[i]);

    g_free(rc);
}

void KeyFile::set_as_int(const std::string &g, const std::string &k, int val)
{
    g_key_file_set_integer(priv->kfile, g.c_str(), k.c_str(), val);
}

void KeyFile::set_as_int_list(const std::string &g, const std::string &k, const std::vector<int> &v)
{
    if (v.empty())
        return;

    gsize length = v.size();
    gint *buf = new int[length];
    for (gsize i=0; i<length; i++)
        buf[i] = v[i];

    g_key_file_set_integer_list(priv->kfile, g.c_str(), k.c_str(),buf, length);
    delete [] buf;
}
