/*
 * data_socket.cc
 *
 *  Created on: Jan 18, 2014
 *      Author: lacasta
 */
#include <iostream>
#include <glib.h>
#include <gio/gio.h>
#include <Plugin.h>
#include <vector>
#include <algorithm>
#include <utils.h>
#include <poll.h>
#ifdef G_OS_WIN32
#include "mingw.h"
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <ifaddrs.h>
#include <netdb.h>
#endif

#define SERVER_PORT 8000

static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading data_socket module...\n");
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading data_socket...\n");
}

std::string get_host_name(struct sockaddr_in *in_addr)
{
    char display[16] = {0};
    inet_ntop(AF_INET, &in_addr->sin_addr.s_addr, display, sizeof display);
    return std::string(display);
}

std::string get_ip_address(const std::string &interface)
{
    std::string out;
    struct ifaddrs *myaddrs, *ifa;
    struct sockaddr_in *s4;
    int status;

    /* buf must be big enough for an IPv6 address (e.g. 3ffe:2fa0:1010:ca22:020a:95ff:fe8a:1cf8) */
    char buf[64];

    status = getifaddrs(&myaddrs);
    if (status != 0)
    {
        err_msg("Error from getifaddrs");
        return out;
    }

    for (ifa = myaddrs; ifa != NULL; ifa = ifa->ifa_next)
    {
        if (ifa->ifa_addr == NULL)
            continue;
        if ((ifa->ifa_flags & IFF_UP) == 0)
            continue;

        if (ifa->ifa_addr->sa_family == AF_INET)
        {
            s4 = (struct sockaddr_in *) (ifa->ifa_addr);
            if (inet_ntop(ifa->ifa_addr->sa_family, (void *) &(s4->sin_addr), buf, sizeof(buf)) == NULL)
            {
                printf("%s: inet_ntop failed!\n", ifa->ifa_name);
            }
            else
            {
                std::cout << "Found " << ifa->ifa_name << "\n\t";
                std::cout << "Checking for interface " << interface << std::endl;
                if (interface.empty() && ( (ifa->ifa_flags & IFF_LOOPBACK)!=0 ))
                {
                    out = buf;
                    break;
                }
                else if (!strcmp(ifa->ifa_name, interface.c_str()))
                {
                    out = buf;
                    break;
                }
            }
        }
    }

    freeifaddrs(myaddrs);
    return out;
}


class DataSocket : public Plugin
{
    private:
        std::vector<int> clients;
        int sckt;
        GThread *thrd;
        GMutex mtx;
        int tmout;

        void clear_socket();
        int send_to_client(int fd, const EventBlock *D);
        static gboolean check_client_status(gpointer);
        static void *accept_connection(DataSocket *);
    public:
        DataSocket();
        ~DataSocket();
        int filter_event(const EventData *data, std::string &S);
        void add_client(int s);
};


DataSocket::DataSocket()
    : sckt(-1), tmout(-1)
{
    int opt = 0, sockfd=0;
    GError *error=0;
    struct sockaddr_in  serv_addr;

    g_mutex_init(&mtx);

    /*
     * Create the socket
     */
    info_msg("DataSocket: creating socket\n");
    sckt = socket(AF_INET, SOCK_STREAM, 0);

    if (sckt < 0 )
    {
        err_msg("DataSocket: error creating socket\n");
        return;
    }

    opt = 1;
    setsockopt (sckt, IPPROTO_TCP, TCP_NODELAY, (char *) &opt, sizeof (opt));
    setsockopt (sckt, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof (opt));

    /*
     * Bind our local address so that the client can send packets to us.
     */
    struct addrinfo hints, *res;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;

    // now get the IP address
    std::string ipaddr = ::get_ip_address("");

    int iok = getaddrinfo(ipaddr.c_str(), 0, &hints, &res);
    if (iok)
    {
        err_msg("Call to getaddrinfo failed with rc=%d\n", iok);
        clear_socket();
        return;
    }
    if (res->ai_next)
    {
        err_msg("Wildcard resolved to multiple addresses\n");
        freeaddrinfo(res);
        clear_socket();
    }
    memcpy(&serv_addr, res->ai_addr, res->ai_addrlen);
    serv_addr.sin_port = htons(SERVER_PORT);
    freeaddrinfo(res);

    if (bind(sckt, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
    {
        err_msg("cannot bind socket");
        clear_socket();
        return;
    }

    std::string xxxx = get_host_name(&serv_addr);
    info_msg("Creating server on %s:%d (%s)\n", ipaddr.c_str(), SERVER_PORT, xxxx.c_str());


    listen(sckt, 5);

    thrd = g_thread_new("accept_connection",(GThreadFunc)accept_connection,this);
    if (error)
    {
        std::cout << "DataSocket: Could not create thread\n\t"
                  << error->message
                  << std::endl;
        g_error_free(error);
        clear_socket();
        return;
    }

    tmout = g_timeout_add(100, (GSourceFunc)check_client_status, this);
}


DataSocket::~DataSocket()
{
    GError *err;
    std::vector<int>::iterator ip;

    if (tmout>=0)
        g_source_remove(tmout);

    g_mutex_lock(&mtx);
    for (ip = clients.begin(); ip!=clients.end(); ip++)
    {
        close(*ip);
    }
    clients.clear();
    g_mutex_unlock(&mtx);
    g_mutex_clear(&mtx);
    g_thread_unref(thrd);
    clear_socket();
}

void DataSocket::clear_socket()
{
    if ( sckt>=0 )
        close(sckt);

    sckt = -1;
}

void DataSocket::add_client(int s)
{
    g_mutex_lock(&mtx);
    clients.push_back(s);
    g_mutex_unlock(&mtx);
}

gboolean DataSocket::check_client_status(gpointer data)
{
    DataSocket *D = (DataSocket *)data;
    std::vector<int>::iterator ip;

    g_mutex_lock(&D->mtx);
    for (ip = D->clients.begin(); ip!=D->clients.end(); ip++)
    {
        struct pollfd PS = { *ip, POLLIN | POLLPRI | POLLERR| POLLHUP, 0 };
        int iok = poll(&PS, 1, 0);
        if ( PS.revents & (POLLHUP | POLLERR) )
        {
            std::cout << "Client " << *ip << " seems to be dead" << std::endl;
            *ip = -1;
        }
    }
    //Remove "dead" clients
    D->clients.erase(std::remove_if(D->clients.begin(),
                                    D->clients.end(),
                                    std::bind2nd(std::less<int>(),0)), D->clients.end()
                     );
    g_mutex_unlock(&D->mtx);
    return TRUE;
}

void *DataSocket::accept_connection( DataSocket *S )
{

    std::cout << "DataSocket: Wait for a connection" << std::endl;
    while (true)
    {
        struct sockaddr_in client_address;
        socklen_t client_len;
        int clientfd = accept( S->sckt, (struct sockaddr *)&client_address, &client_len);
        if (clientfd<0)
        {
            err_msg("Error accepting socket\n");
            return 0;
        }
        std::cout << "DataSocket: Got a new client connection" << std::endl;
        S->add_client(clientfd);
    }
    return 0;
}


int DataSocket::filter_event(const EventData *data, std::string &S)
{
    if (!sckt)
        return 0;

    int rc;
    const EventBlock *B = data->get_block();
    std::vector<int>::iterator ip;
    g_mutex_lock(&mtx);
    for (ip = clients.begin(); ip!=clients.end(); ip++)
    {
        int iok;
        iok = send_to_client(*ip, B);
        if (iok<0)
        {
            close(*ip);
            *ip = -1;
        }
        if (iok>0)
            rc += iok;
    }
    //Remove "dead" clients
    clients.erase(std::remove_if(clients.begin(),clients.end(), std::bind2nd(std::less<int>(),0)), clients.end());
    g_mutex_unlock(&mtx);
    return rc;
}

int DataSocket::send_to_client(int fd, const EventBlock *B)
{
    guint32 iword;
    int sent = 0;
    int iok;
    iword = 0xcafe0000 | 256;

    iok = send(fd, (const gchar *)&iword, sizeof(guint32), 0);
    if (iok<0)
    {
        if (errno == EPIPE )
            err_msg("Client dead!!!\n");
        else
            err_msg( "Error sending header\n");
        return -1;
    }
    else
        sent += iok;


    for (int ichip=0; ichip<2; ichip++)
    {
        const char *buff = (const char *)B->chip[ichip].signal;
        size_t out, sz = 128*sizeof(unsigned short);
        for (out=0;out<sz;)
        {
            iok = send(fd, buff, 128*sizeof(short), 0);
            if (iok<0)
            {
                err_msg("Error sending data from chip 0.\n");
                return -2;
            }
            else
                out += iok;
        }
        sent += out;
    }
    return sent;
}

extern "C"
{
    Plugin *create_plugin()
    {
        DataSocket *plugin = new DataSocket;
        return plugin;
    }
}
